const DEBUG = false;

require('./src/init');

const {
    TYPE_PROPERTY,
    TYPE_EVENT,
    TYPE_SERVICE
} = require('./src/const');

const {
    startServer
} = require('./src/socket/server');

const {
    parse
} = require('./src/parser');

let sdkType = 'linkedge-thing-access-sdk';
if (DEBUG) {
    sdkType = './src/mock/ta-sdk.js';
}
const {
    ThingAccessClient,
    Config
} = require(sdkType);

const stateMap = new Map();
const clients = new Map();

async function main() {
    // Parse driver config to get connection info.
    const config = await Config.get();
    config.getThingInfos().forEach((thingInfo) => {
        clients.set(getCloudId(thingInfo.productKey, thingInfo.deviceName), bind(thingInfo));
    });

    startServer((data, socket) => {
        function replyClient(code, object) {
            console.log(`replyClient: ${code}, ${object}`);
            if (object instanceof Error) {
                object = `${object}`;
            }
            let payload = {
                code,
                message: object
            };
            if (code == -1) {
                payload = object;
            }
            socket.write(`${JSON.stringify(payload)}`);
            console.log(`reply: ${JSON.stringify(payload)}`);
        }
        let result;
        try {
            result = parse(data);
        } catch (e) {
            replyClient(1, `Parse failed: ${e}`);
            return;
        }
        if (!result) {
            replyClient(1, `Parse failed: result is null!`);
            return;
        }
        const cloudId = getCloudId(result.productKey, result.deviceName);
        const client = clients.get(cloudId);
        if (client == null) {
            replyClient(1, `Can not found device[productKey=${result.productKey},deviceName=${result.deviceName}]! Please register device first!`)
            return;
        }
        switch (result.type) {
            case TYPE_PROPERTY:
                _setProperties(cloudId, result.params);
                replyClient(0, 'Send property success!');
                break;
            case TYPE_EVENT:
                client.reportEvent(result.identifier, result.params);
                replyClient(0, 'Send event success!');
                break;
            case TYPE_SERVICE:
                if (!DEBUG) {
                    replyClient(1, new Error(`Not support type: ${result.type}`));
                    break;
                }
                let object;
                try {
                    object = client.mockCallServices(result.identifier, result.params);
                } catch (e) {
                    replyClient(1, e);
                    break;
                }
                if (object instanceof Promise) {
                    object.then((ret) => {
                        replyClient(-1, ret);
                    }).catch(e => replyClient(1, e));
                } else {
                    replyClient(-1, object);
                }
                break;
            default:
                replyClient(1, new Error(`Not support type: ${result.type}`));
                break;
        }
    });
}

function getServicesHandler(cloudId) {
    return {
        setProperties: function (properties) {
            // Set properties to thing and return the result.
            // Return a object result or a promise of it. Throw the error if one raises.
            _setProperties(cloudId, properties);
            return {
                code: 0,
                message: 'success'
            }
        },
        getProperties: function (keys) {
            // Get properties from thing and return them.
            // Return a object result or a promise of it. Throw the error if one raises.
            return {
                code: 0,
                message: 'success',
                params: _getProperties(cloudId, keys)
            };
        },
        callService: function (name, args) {
            // Call services on thing and return the result.
            // Return a object result or a promise of it. Throw the error if one raises.
            return new Promise((resolve, reject) => {
                clients.get(cloudId).getTsl().then((tslStr) => {
                    const services = JSON.parse(tslStr).services;
                    if (!existService(services, name)) {
                        reject(new Error(`No support service[${name}]!`));
                        return;
                    }
                    console.log(`Device[${cloudId}] callService[${name}] with ${JSON.stringify(args)}`);
                    resolve({
                        code: 0,
                        message: 'success'
                    });
                }).catch(e => {
                    reject(new Error(`Parse TSL failed! ${e}`));
                });
            });
        }
    };
}

function existService(services, serviceName) {
    for (let i = 0; i < services.length; i++) {
        if (services[i].identifier === serviceName) {
            return true;
        }
    }
    return false;
}

function getCloudId(productKey, deviceName) {
    return `${productKey}_${deviceName}`;
}

function bind(config) {
    let client = new ThingAccessClient(config, getServicesHandler(getCloudId(config.productKey, config.deviceName)));
    client.setup()
        .then(() => {
            return client.registerAndOnline();
        })
        .catch(err => {
            client.cleanup();
            console.log(err);
        })
        .catch(err => {
            console.log(err);
        });
    return client;
}

function _setProperties(cloudId, properties) {
    if (!stateMap.has(cloudId)) {
        stateMap.set(cloudId, {});
    }
    stateMap.set(cloudId, Object.assign(stateMap.get(cloudId), properties));
    clients.get(cloudId).reportProperties(properties);
}

function _getProperties(cloudId, keys) {
    if (!stateMap.has(cloudId)) {
        return {};
    }
    const result = {};
    keys.forEach(key => {
        result[key] = stateMap.get(cloudId)[key];
    });
    return result;
}

main();

// This is a handler which never be invoked in the example.
module.exports.handler = function (event, context, callback) {
    callback(null);
};